""" Propagators for classical particle dynamics """

import numpy as np

class Propagator:
    """ Base class for all propagators for classical particle dynamics """
    name = 'generic'
    def __init__(self, syst, tstart, tend, tstep, substep=None, pbc=False,
                 tqdm=False):
        self.q = syst.q0
        self.p = syst.p0
        self.q0 = syst.q0
        self.p0 = syst.p0
        self.syst = syst
        self.tstart = tstart
        self.tend = tend
        self.tstep = tstep
        self.substep = substep
        self.en0 = self.syst.energy(self.q0, self.p0)
        assert self.tstart < self.tend
        nsteps = int(np.floor((self.tend-self.tstart)/self.tstep))+1
        # self.time = np.linspace(self.tstart, self.tend, nsteps)
        self.time = [tstart+(tend-tstart)/(nsteps-1)*s for s in range(nsteps)]
        self.sq = [self.q]
        self.sp = [self.p]
        self.pbc = pbc
        self.tqdm = tqdm

        if self.substep is not None:
            assert self.substep < self.tstep
            assert np.isclose(self.tstep % self.substep, 0)
            nsubsteps = int(np.floor(self.tstep/self.substep))
            self.subtime = np.linspace(self.substep, self.tstep, nsubsteps)

    def propagate(self):
        """ perform propagation of coordinate and momentum in time """
        from tqdm import tqdm
        (q, p) = (self.q, self.p)
        if self.substep is not None:
            its = tqdm(self.time[:-1]) if self.tqdm else self.time[:-1]
            for time in its:
                (qt, pt) = self.step(time, q, p)
                self.sq.extend(qt)
                self.sp.extend(pt)
                q = qt[-1]
                p = pt[-1]
            self.time = [t+st for t in self.time[:-1] for st in self.subtime]
            self.time.insert(0, 0)
        else:
            its = tqdm(self.time[1:], miniters=10) if self.tqdm else self.time[1:]
            for time in its:
                (q, p) = self.step(time, q, p)
                self.sq.append(q)
                self.sp.append(p)

    def step(self, time, coordinate, momentum):
        """ perform a time step - stub method """
        return (coordinate, momentum)

    def get_trajectory(self, step=1, dtype=np.float):
        """ return a trajectory - a time series of coordinate and momentum """
        time = np.array(self.time, dtype=dtype)[::step]
        coordinate = np.array(self.sq, dtype=dtype)[::step]
        momentum = np.array(self.sp, dtype=dtype)[::step]
        return time, coordinate, momentum

    def get_trajectory_3d(self, step=1):
        """ return a trajectory of coordinates grouped in 3-tuples """
        ti, qt, pt = self.get_trajectory(step)
        qt.shape = (len(qt), -1, 3)
        pt.shape = (len(pt), -1, 3)
        return ti, qt, pt

    def analyse(self, step=1, p='err'):
        """ calculate relative energy error or relative energy difference
            (p = 1, 2, 3, ..., inf), see refs. https://doi.org/10.2307/2683905
            https://en.wikipedia.org/wiki/Relative_change_and_difference """
        from itertools import islice
        if not hasattr(self, 'en'):
            qptup = islice(zip(self.sq, self.sp), 0, None, step)
            self.en = np.array([self.syst.energy(*t) for t in qptup])
        if p == 'inf':
            func = lambda x, y: (x-y)/np.maximum(abs(x), abs(y))
        elif p == 'err':
            func = lambda x, y: (x-y)/y
        elif isinstance(p, int) and p > 0:
            func = lambda x, y: (x-y)/(abs(x)**p+abs(y)**p)**(1/p)
        else:
            raise ValueError('invalid value for p: '+repr(p))
        self.er = np.array(func(self.en, self.en0), dtype=float)


from liouville import LiouvillePowersQP

class PolynomialPropagator(Propagator):
    """ Base class for polynomial classical propagators """
    name = 'polynomial'
    def __init__(self, liouville=LiouvillePowersQP, liou_params={},
                 restart=False, restart_file='propag.pkl', fpprec='fixed',
                 prec=None, **kwargs):
        super().__init__(**kwargs)
        self.step = self.step_pbc if self.pbc else self.step_nopbc
        self.liouville = liouville
        self.liou_params = liou_params
        self.restart = restart
        self.restart_file = restart_file
        self.fpprec = fpprec
        self.prec = prec

    def set_polynomial(self, **kwargs):
        """ initialize the polynomial (iln) object """
        if self.restart:
            self.iln = self.load(self.restart_file)
        else:
            params = {'fpprec': self.fpprec, **kwargs, **self.liou_params}
            self.iln = self.liouville(self.nterms, **params)

    def step_nopbc(self, time, q, p):
        """ perform a time step with initial coordinate and momentum """
        qvec, pvec = self.iln.get_val_float(q, p)
        qt = np.dot(self.coeffs, qvec)
        pt = np.dot(self.coeffs, pvec)
        return (qt, pt)

    def step_pbc(self, time, q, p):
        """ perform a time step with periodic boundary conditions """
        qvec, pvec = self.iln.get_val_float(q, p)
        qt = np.dot(self.coeffs, qvec)
        pt = np.dot(self.coeffs, pvec)
        return (self.syst.wrap_q(qt), pt)

    def load(self, filename='propag.pkl'):
        """ deserialize the propagator object and extract the iln object """
        import cloudpickle as pkl
        with open(filename, 'rb') as fh:
            that = pkl.load(fh)
        params = ['__class__', 'name', 'nterms', 'delta', 'tstep', 'scale',
                  'shift', 'liouville', 'liou_params', 'fpprec', 'prec']
        for key in params:
            attr = getattr(that, key)
            errstr = key + ': ' + str(getattr(self, key)) + ' != ' + str(attr)
            assert getattr(self, key) == attr, errstr
        assert self.syst.hamiltonian == that.syst.hamiltonian
        return that.iln

    def dump(self, filename='propag.pkl'):
        """ serialize and save the propagator object """
        import cloudpickle as pkl
        with open(filename, 'wb') as fh:
            pkl.dump(self, fh)
